// Marcin Miszczak G1ISI
// Temat pracy domowej: Dziedziczenie wielobazowe i wirtualne

#include <iostream>
#include <cstdlib>

using namespace std;

/* -----------------------------------------------------------------------------------------

   Czym jest dziedziczenie wielobazowe (MI)?
      Dziedziecznie wielobazowe (MI) oznacza mozliwosc dziedziczenia po wiecej niz jednej
      klasie bazowej tzn:
  
         class Derived : public A, public B {...};
   
      W tym przypadku instancja klasy Derived "JEST" zarowno instancja klasy A jak i B tj.
      dozwolone sa rzutowania z klasy Derived na A oraz B

   Jakie sa uzasadnione przypadki korzystania z MI?
      #1 Laczenie bibliotek i modulow  - rozszerzanie funkcjonalnosci bibliotek, z ktorych kazda wymaga
         dziedziczenia po roznych klasach
	  #2 Latwosc polimorficznego uzycia - obiekt wielobazowy mozemy przekazywac do metod, ktore spodziewaja
	     sie obiektow klas bazowych. Uzywamy jednego obiektu pochodnego zamiast kilku obiektow klas bazowych
	  #3 Uzyskanie jednej okreslonej hierarchii klas - dla przykladu std::exception

   Czy moge korzystac z zalet MI nie uzywajac MI?
      Tak! Mozna skorzystac z agregacji jesli narzut zwiazny z dodatkowym wywolaniem jest mniejszy niz
      narzut zwiazany z dziedziczeniem wielobazowym

   Do czego sluzy mechanizm dziedziczenia wirtualnego?
      Dziedziczenia wirtualnego uzywany przede wszystkim do tego aby wieliminowac mozliwosc wystepowania
      kilku instancji tych samych klas w hierarchii dziedziczenia. Aby wyrazic dziedziczenie wirtualne
      nalezy przy specyfikowaniu operacji dziedziczenia uzyc slowa kluczowego virtual w miejscu, w ktorym 
	  wiemy, ze dany obiekt zostanie wlaczony do hierarchii dziedziczenia wiecej niz jeden raz

   ----------------------------------------------------------------------------------------- */

// Dziedziczenie wielobazowe (MI) - przyklad 0x00 - elementarny przyklad dziedziczenia
class A // klasa polozona najwyzej w hierarchii dziedziczenia
{
	public:
		A(int a = 1) : a_(a) {cout << "A " << this << endl;} // uzywane aby pokazac kolejnosc wolania konstruktorow
		~A() {cout << "~A" << endl;} // uzywane aby pokazac kolejnosc wolania destruktorow
		int a_; // uzywane aby pokazac niejednoznacznosc odwolania wynikajaca z MI
};

class B : public A
{
	public:
		B(int b) : A(b) {cout << "B " << this << endl;} 
		~B() {cout << "~B" << endl;}
};

class C : public A
{
	public:
		C(int c) : A(c) {cout << "C " << this << endl;}
		~C() {cout << "~C" << endl;}
};

class MI : public B, public C // klasa MI dziedziczy wielobazowo z klas B i C, ktore dziedzicza z A
{
	public:
		MI(int i = 1) : B(i), C(i) {cout << "MI " << this << endl;}
		~MI() {cout << "~MI" << endl;}
		// test odwolania do zmiennych, ktore wystepuja wiele razy w hierarchii dziedziczenia
		void getVars(){ cout << "B::a_ = " << B::a_ << " C::a_ = " << C::a_ << endl;}
};

// Dziedziczenie  wirtualne - przyklad 0x01 - "delegation to a sister class"
class VA // klasa polozona najwyzej w hierarchii dziedziczenia, klasa czysto abstrakcyjna
{
	public:
		VA(int a = 1) : a_(a) {cout << "VA " << this << endl;} // uzywane aby pokazac kolejnosc wolania konstruktorow
		virtual ~VA() {cout << "~VA" << endl;} // uzywane aby pokazac kolejnosc wolania destruktorow
		virtual void foo() = 0; // uzywamy po to aby pokazac dzialanie mechanizmu
		virtual void bar() = 0; // "delegation to a sister class"
		int a_; // odwolanie do a_ jest juz jednoznaczne
};

class VB : virtual public VA // klasa VB dziedziczy wirtualnie z klasy VA - w tym miejscu nalezy podac s.k. virtual
{
	public:
		VB(int b) : VA(b) {cout << "VB " << this << endl;} 
		~VB() {cout << "~VB" << endl;}
		// definicja metody foo(), kompilator umieszcza odpowiedni adres w wspoldzielonej tablicy VTABLE
		// przy czym celowo nie definiujemy metody bar()
		// UWAGA! Kompilator moze generowac warningi typu: inherits 'VB::VB::foo' via dominance
		void foo() {cout << "Wywoluje bar() z klasy VB " << endl; bar();}
};

class VC : virtual public VA // klasa VC dziedziczy wirtualnie z klasy VA
{
	public:
		VC(int c) : VA(c) {cout << "VC " << this << endl;}
		~VC() {cout << "~VC" << endl;}
		// definicja metody bar(), kompilator umieszcza odpowiedni adres w wspoldzielonej tablicy VTABLE
		// przy czym celowo nie definiujemy metody foo()
		// UWAGA! Kompilator moze generowac warningi typu: inherits 'VC::VC::foo' via dominance
		void bar() {cout << "Wywolalem bar() z klasy VC !" << endl;}
};

class VMI : public VB, public VC // klasa VMI dziedziczy wielobazowo z klas VB i VC, ktore dziedzicza wirtualnie z A. 
{
	public:
		VMI(int i = 1) : VB(i), VC(i) {cout << "VMI " << this << endl;}
		~VMI() {cout << "~VMI" << endl;}
};

// Dziedziczenie wielobazowe - przyklad 0x02 - wykorzystanie jednego obiektu jako argumentu dla kilku funkcji
void funcb(const B*){ cout << "Witamy w funcb(const B*)" << endl; }
void funcc(const C*){ cout << "Witamy w funcc(const C*)" << endl; }

int main(int argc, char* argv[])
{

/* -----------------------------------------------------------------------------------------

	   Hierarchia klas w przykladzie 0x00 jest nastepujaca:
       A  A    Widzimy, ze obiekt klasy MI zawiera po jednym obiekcie klas B i C, mozemy
	   |  |    takze zobaczyc, ze w h. dziedziczenia obiekt klasy A wystepuje az 2 razy!
       B  C    Z tego wzgledu ponizszy kod (2 liniki) nie kompiluje sie poprawnie, a sam
	    \/     kompilator (VC) alarmuje: ambiguous access of 'a_' ... Nie oznacza to, ze
		MI     programista nie jest w stanie odwolywac sie do takiej zmiennej a_ - moze
		       za pomoca nastepujacego odwolania wewnatrz metody:

			   nadklasa::nazwa_zmiennej - test odwolania (metoda getVars())

			   Taka hierarchia dziecziczenia okreslana jest mianem "dreaded diamond"

			   Q: Jaka jest kolejnosc wywolywania konstruktorow?
			   A: A, B, A, C, MI - wynikiem takiej a nie innej sekwencji jest kolejnosc 
			      wystepowania nazw klas na liscie klas z ktorych dziedziczymy, kolejnosc
				  wystepowanie nazw klas na liscie inicjalizacyjnej konstr. jest ignorowana
			   Q: Jaka jest kolejnosc wywolywania destruktorow ?
			   A: Kolejnosc jest odwrotna: MI, C, A, B, A
   ----------------------------------------------------------------------------------------- */
	{
		MI mi;
		//mi.a_ = 2; //ambiguous access of 'a_' could be the 'a_' in base 'A' or could be the 'a_' in base 'A'
		
		B* b = static_cast<B*>(&mi); // bezpiecznie rzutowanie z klasy pochodznej na jedna z klas bazowych
		b->a_ = 37;
		
		C* c = static_cast<C*>(&mi); // jw.
		c->a_ = 39;
		
		mi.getVars(); // test sprawdzajacy odwolywanie sie do zmiennych wystepujacych w h. dz. wielokrotnie
	} // sumulacja kolejnosci wywolywania destruktorow
	
	MI mi;
	funcb(static_cast<const B*>(&mi));
	funcc(static_cast<const C*>(&mi));
	// obiekt mi moze zostac wielokrotnie wykorzystany jako argument dla kilku roznych funkcji
	
	{

/* -----------------------------------------------------------------------------------------

	   Hierarchia klas w przykladzie 0x01 jest nastepujaca:
        VA     Widzimy, ze obiekt klasy VMI zawiera po jednym obiekcie klas VB i VC oraz
	   / \     w przeciwiensewie do MI obiekt klasy VA wystepuje dokladnie jeden raz!
      VB  VC   Tym razem odwolanie do zmiennej a_ jest juz jednoznaczne.
	    \/     
		MI     Q: Jaka jest kolejnosc wywolywania konstruktorow?
			   A: A, VB, VC, VMI - przyczyna pojawienia sie A jako pierwszego wolanego kostruktora
			      NIE jest fakt, ze wystepuje on najwyzej w hierarchii dziedziczenia ale fakt, ze
				  jest on dziedziczony wirtualnie! Kolejnosc wywolywania konstruktorow jest zgodna z
				  zasada "depth-first left-to-right traversal of the graph of base classes" przy czym 
				  najpierw brane sa pod uwage klasy dziedziczone wirtualnie. Konsekwencja tego faktu 
				  jest koniecznosc zadeklarowania konstruktora klasy najwyzej polozoej w h jako
				  domyslnego lub takiego, ktory ma zadeklarowane wszystkie wartosci jako domyslne
			   Q: Jaka jest kolejnosc wywolywania destruktorow ?
			   A: Kolejnosc jest odwrotna: VMI, VC, VB, A

   ----------------------------------------------------------------------------------------- */

		VMI vmi;
		vmi.a_ = 2; // odwolanie jest juz jednoznaczne

		vmi.foo(); /* wywolanie pokazuje w jaki sposob mozna wywolac metode klasy VC z klasy VB! Przy czym
				      tak naprawde klasa VB nie ma nic wspolnego z klasa VC oprocz tego, ze posiadaja
					  wskazanie VPTR do tej samej tablicy VTABLE - co uzasadnia owe specyficzne dzialanie
					  Sytuacje taka okresla sie mianem "delegation to a sister class" */
	} // sumulacja kolejnosci wywolywania destruktorow
	
	return 0;
}
